// a simple file writing test.. will try to gzip and md5sum the output in real time
//
// to compile:
// 
//  g++ -Wall filewrite.cpp -o filewrite -lcrypto -lz

// NOTE, will just take the output block and use that as the compression chunk size
// so small amounts of data will not compress well.
// send bigger blocks if you have a choice (buffer up some shots or something first)

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <errno.h>

#include <string.h>   // required for memcpy()
#include <sys/time.h> // required for gettimeofday()
#include <time.h>     // required for struct tm
#include <unistd.h>   // required for sleep()

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/statvfs.h>

#include "filewrite.h"

using namespace std;

// constructor
lvisf_file_write::lvisf_file_write()
//: theDay(-1)
//,theMonth(-1)
//, theYear(-1)
{
   init();  // just init ourselves in case the fool running this did not
}

// destructor
lvisf_file_write::~lvisf_file_write()
{
   //printf("filewrite: (destructor) Closing current file.\n");
   //file_close( );
}

void lvisf_file_write::init()
{
   int i;

   if( m_flag_systemInitialized == false )
     {
	theYear = -1;
	theMonth = -1;
	theDay = -1;
	memset(m_char_extension,            '\0',sizeof(m_char_extension));
	memset(m_char_filename,             '\0',sizeof(m_char_filename));
	memset(m_char_basepath_a,           '\0',sizeof(m_char_basepath_a));
	memset(m_char_basepath_b,           '\0',sizeof(m_char_basepath_b));
	memset(m_char_filename_a,           '\0',sizeof(m_char_filename_a));
	memset(m_char_filename_b,           '\0',sizeof(m_char_filename_b));
	memset(m_char_filename_md5sum,      '\0',sizeof(m_char_filename_md5sum));
	memset(m_char_filenameUncompressed, '\0',sizeof(m_char_filenameUncompressed));
	strncpy(m_char_filename,  LVISF_FILENAME_NOT_SET,  sizeof(LVISF_FILENAME_NOT_SET));
	strncpy(m_char_basepath_a,LVISF_DATA_BASEPATH,     sizeof(LVISF_DATA_BASEPATH));
	strncpy(m_char_basepath_b,LVISF_FILENAME_NONE,     sizeof(LVISF_FILENAME_NONE));
	strncpy(m_char_extension, LVISF_DATA_EXTENSION,    sizeof(LVISF_DATA_EXTENSION));
	strncpy(m_char_md5sumkernel, LVISF_MD5SUM_DEFAULT, sizeof(LVISF_MD5SUM_DEFAULT));
	m_counter_blockNumber = 0;
	m_counter_bytesWritten = 0;
	m_counter_fileNumber = 1;
	m_counter_currentBuffer = m_counter_lastBuffer = 0;
	m_flag_compressFile = false;        // should we be compressing file?
	m_flag_fileOpen = false;            // do we have a file currently open?
	m_flag_hk_fileOpen = false;
	m_flag_gzipReady = false;           // by default, no, gzip is not ready
	m_flag_md5sumFile = true;          // calculate md5sum as we write?
	m_flag_openNewFile = false;         // set true when time to open a new file
	m_flag_secondFileEnabled = false;   // by default we are only writing to one file
	m_flag_primaryFileEnabled = true;   // by default we are only writing to one file
	//   m_limit_bytes = 500 * 1024 * 1024;  // 500 MB default file size (old lvis default)
	m_limit_bytes = 2.5 * 1024 * 1024 * 1024;  // 500 MB default file size (old lvis default)
	m_limit_bytesPerWrite = 1024 * 1024;// 1 MB per write
	m_limit_delayPerWrite = 1000;       // 1 millisecond per write
	m_limit_seconds = 5 * 60;           // 5 minutes per file if size isn't breached
	m_time_fileOpenSeconds = 0;
	m_fp  = NULL;
	m_fpa = NULL;
	m_fpb = NULL;
	
	for(i=0;i<LVISF_RAW_DATA_BUFFERS;i++)
	  m_data_out[i] = new unsigned char[LVISF_RAW_DATA_SIZE];
	
	//printf("init - theDay [%d] [%d] [%d]\n",theYear, theMonth, theDay);
	
	//printf("filewrite: (constructor)  theDay [%d]\n",theDay);
	
	//
	// Turn off cache. Do this here, open() is called on every new file!
	// 
	// system("hdparm -W 0 /dev/sdb1");  // /dev/sdb1 for us
   	// 
	////printf("filewrite: (constructor) Allocated Write buffer...[%d buffers of %d bytes each]\n", LVISF_RAW_DATA_BUFFERS, LVISF_RAW_DATA_SIZE );
     }
 
   m_flag_systemInitialized  = true;
}

int lvisf_file_write::calc_file_md5sum( char * filename, unsigned char * md5result )
{
   
   unsigned char c[MD5_DIGEST_LENGTH];
   int i;
   FILE *inFile = fopen(filename, "rb");
   MD5_CTX mdContext;
   int bytes;
   unsigned char data[64*1024];
   
   if (inFile == NULL) {
      // printf ("%s can't be opened.\n", filename);
      return -1;
   }
   
   MD5_Init (&mdContext);
   while ((bytes = fread (data, 1, sizeof(data), inFile)) != 0)
     MD5_Update (&mdContext, data, bytes);
   MD5_Final (c,&mdContext);
   for(i = 0; i < MD5_DIGEST_LENGTH; i++)
     {
	// printf("%02x", c[i]);
	md5result[i] = c[i];
     }
   // printf (" %s\n", filename);
   fclose (inFile);
   return (int) MD5_DIGEST_LENGTH;  
}

int lvisf_file_write::get_file_modtime( char * filename, char * datestring )
{
   int retval;
   struct stat attrib;
   retval = stat(filename, &attrib);
   if(retval == 0)
     {	
	strftime(datestring, 20, "%Y-%m-%d_%H:%M:%S", localtime(&(attrib.st_ctime)));
     }
   
   return retval;
}

void lvisf_file_write::print_theDay()
{
   printf("theDay [%d]\n",theDay);
}

void lvisf_file_write::file_close( )
{
   int i;
   //size_t bytesFromFwrite;
   
   // close the file pointer if already open
   if(NULL != m_fpa)
     {
	if( true == m_flag_md5sumFile ) // close out md5sum if we are doing that
	  MD5_Final ( m_md5sum , &m_mdContext );	     

/*
	if( true == m_flag_compressFile ) // flush out the gzip buffer (end the file)
	  {
	     m_strm.avail_out = sizeof ( unsigned char) * LVISF_RAW_DATA_SIZE;
	     m_strm.next_out = m_data_out[ m_counter_currentBuffer ] ;
	     m_strm.avail_in = 0;  // no new data, just finish the stream
	     ret = deflate( &m_strm, Z_FINISH );
	     m_counter_gzipBytesOut = sizeof ( unsigned char) * LVISF_RAW_DATA_SIZE - m_strm.avail_out;
	     bytesFromFwrite  = fwrite( m_data_out [ m_counter_currentBuffer ] , 1, m_counter_gzipBytesOut, m_fpa);
	     if( true == m_flag_secondFileEnabled)
	       if( NULL != m_fpb )
		 {
		    fwrite( m_data_out [ m_counter_currentBuffer ], 1, m_counter_gzipBytesOut, m_fpb);		    
		 }
	     
	     //fprintf(stdout,"filewrite: (file_close) Wrote out %d bytes.\n", ( int ) bytesFromFwrite);
	     deflateEnd (& m_strm);
	  }
*/

	m_flag_fileOpen = false;
	m_flag_hk_fileOpen = false;
	fclose(m_fpa);
	printf("Closed file %s\n", m_char_filename_a); 
	
	m_fpa = NULL;
	
	if( true == m_flag_secondFileEnabled)
	  if( NULL != m_fpb )
	    {
	       fclose(m_fpb);
	       
	       printf("Closed file %s\n", m_char_filename_b);
	       m_fpb = NULL;
	    }
	
	if( true == m_flag_md5sumFile)
	  {
	     fprintf(stdout,"md5sum of last file = ");
	     for( i = 0; i < MD5_DIGEST_LENGTH; i++ )
	       fprintf(stdout,"%02x", m_md5sum[i] );
	     fprintf(stdout,"\n");
	     
	     // append to "md5sum.txt"
	     // generate a new file name
	     sprintf( m_char_filename_md5sum, "%s%s", m_char_basepath_a, m_char_md5sumkernel);	     
	     m_fp = fopen(m_char_filename_md5sum,"ab");
	     if(NULL != m_fp)
	       {		 
		  fprintf(m_fp,"%10lu  ", (unsigned long) m_counter_bytesWritten );
		  for( i = 0; i < MD5_DIGEST_LENGTH; i++ )
		    fprintf(m_fp,"%02x", m_md5sum[i] );
		  fprintf(m_fp,"  %s\n", m_char_filenameUncompressed );
		  fclose(m_fp);
		   
	       }	
	     
	     if( true == m_flag_secondFileEnabled)
	       {
		  sprintf( m_char_filename_md5sum, "%s%s", m_char_basepath_b, m_char_md5sumkernel);	     
		  m_fp = fopen(m_char_filename_md5sum,"ab");
		  if(NULL != m_fp)
		    {		  
		       fprintf(m_fp,"%10lu  ", (unsigned long) m_counter_bytesWritten );
		       for( i = 0; i < MD5_DIGEST_LENGTH; i++ )
			 fprintf(m_fp,"%02x", m_md5sum[i] );
		       fprintf(m_fp,"  %s\n", m_char_filenameUncompressed );
		       fclose(m_fp);
		    }	
	       }	     
	  }	
     }  
}

void lvisf_file_write::file_flushCache( )
{
   if( NULL != m_fp )
     {		  
	fflush( m_fp );
     }

   if( NULL != m_fpa )
     {		  
	fflush( m_fpa );
     }

   if( true == m_flag_secondFileEnabled )
     {
	if( NULL != m_fpb )
	  {
	     fflush( m_fpb );
	  }
     }
}

int lvisf_file_write::file_open(bool useBaseFilename)
{
   struct tm myTm;
   time_t local_time;
   
   file_close( );
   
   if(useBaseFilename == false)
   {
        memset(&myTm,0,sizeof(myTm));  // zero this out

        //
        // Need this all the time for hours,mins,sec
        //
        time(&local_time);
        memcpy(&myTm,localtime(&local_time),sizeof(struct tm));

        //
        // Set the day value upon first file open
        //
        if (theDay < 0)
            {
                theYear = myTm.tm_year + 1900;
                theMonth = myTm.tm_mon + 1;
                theDay = myTm.tm_mday;
            }
        else
            {
                if(theDay != myTm.tm_mday) myTm.tm_hour += 24; // add a day of hours if we've rolled over
            }

        //  
        // generate a new file name (first with no path for the md5sum name)
        // theDay value is used to keep the day constant, dlr does not want the day to change
        //
        sprintf( m_char_filename_a, "%s_%04d%02d%02d_%02d%02d%02d_%8.8d", m_char_filename,
                theYear , theMonth , theDay ,
                myTm.tm_hour , myTm.tm_min , myTm.tm_sec ,
                m_counter_fileNumber );   
        strncpy( m_char_filenameUncompressed, m_char_filename_a, sizeof(m_char_filenameUncompressed) );  // copy this file base to here

        sprintf( m_char_filename_a, "%s%s_%04d%02d%02d_%02d%02d%02d_%8.8d", m_char_basepath_a, m_char_filename,
                theYear , theMonth , theDay ,
                myTm.tm_hour , myTm.tm_min , myTm.tm_sec ,
                m_counter_fileNumber );
        strcat( m_char_filename_a, m_char_extension );            // add the proper extension

        strcat( m_char_filenameUncompressed, LVISF_DATA_EXTENSION );  // use the uncompressed extension for md5sum filename
   }
   else
   {
       sprintf( m_char_filename_a, "%s%s", m_char_basepath_a, m_char_filename);
   }

   m_fpa = fopen(m_char_filename_a,"ab");
   //// m_fid = open(m_char_filename_a,O_CREAT|O_WRONLY|O_NONBLOCK, 0666);
   //// fprintf(stdout,"filewrite: (file_open) Opening a new file (A): %s return [%d]\n", m_char_filename_a ,m_fid);
   fprintf(stdout,"filewrite: (file_open) Opening a new file (A): %s\n", m_char_filename_a );

   if( (true == m_flag_secondFileEnabled) && ( strncmp(m_char_basepath_b,"NONE",4)!=0 ) )
   {
      if(useBaseFilename == false)
      {
        // generate a new file name (first with no path for the md5sum name)
        sprintf( m_char_filename_b, "%s_%04d%02d%02d_%02d%02d%02d_%8.8d", m_char_filename,
            theYear , theMonth , theDay ,
                    myTm.tm_hour , myTm.tm_min , myTm.tm_sec ,
                    m_counter_fileNumber );
        strncpy( m_char_filenameUncompressed, m_char_filename_b, sizeof(m_char_filenameUncompressed) );  // copy this file base to here
        sprintf( m_char_filename_b, "%s%s_%04d%02d%02d_%02d%02d%02d_%8.8d", m_char_basepath_b, m_char_filename,
            theYear , theMonth , theDay ,
                    myTm.tm_hour , myTm.tm_min , myTm.tm_sec ,
                    m_counter_fileNumber );
        strcat( m_char_filename_b, m_char_extension );            // add the proper extension
        strcat( m_char_filenameUncompressed, LVISF_DATA_EXTENSION );  // use the uncompressed extension for md5sum filename

        fprintf(stdout,"filewrite: (file_open) Opening a new file (B): %s\n", m_char_filename_b );
      }
      else
      {
        sprintf( m_char_filename_b, "%s%s", m_char_basepath_b, m_char_filename);
      }  

      m_fpb = fopen(m_char_filename_b,"ab");
      if ( NULL == m_fpb )
      { 
        fprintf(stderr,"ERROR: Opening file %s \n",m_char_filename_b);
      }
   }
   
   if ( NULL == m_fpa )  // we had a problem opening the file!?
   {
      printf("Error opening %s\n",m_char_filename_a);
      return 1;
   }
   else               // OK to proceed, reset some variables
   {
      m_counter_fileNumber++;
      m_flag_fileOpen = true;
      m_counter_bytesWritten = 0;  // clear bytes written
      m_time_fileOpenSeconds = time_getEpochSeconds( );
      if( true == m_flag_md5sumFile) MD5_Init(&m_mdContext);
      //if( true == m_flag_compressFile) gzip_init( );
      return 0;
   }
}

int lvisf_file_write::file_hk_open(void)
{
   struct tm myTm;
   time_t local_time;
   
   file_close( );
         
   //
   // Need this all the time for hours,mins,sec
   //
   time(&local_time);
   memcpy(&myTm,localtime(&local_time),sizeof(struct tm));
   
   //
   // Set the day value upon first file open
   //
   if (theDay < 0)
   {
      theYear = myTm.tm_year + 1900;
      theMonth = myTm.tm_mon + 1;
      theDay = myTm.tm_mday;
   }
        
   //  
   // generate a new file name (first with no path for the md5sum name)
   // theDay value is used to keep the day constant, dlr does not want the day to change
   //
   sprintf( m_char_filename_a, "%s%s_%04d%02d%02d_%02d%02d%02d_%8.8d", m_char_basepath_a, m_char_filename,
	theYear , theMonth , theDay ,
	myTm.tm_hour , myTm.tm_min , myTm.tm_sec ,
	m_counter_fileNumber );
   strcat( m_char_filename_a, ".csv" );            // add the proper extension
   
   strcat( m_char_filenameUncompressed, LVISF_DATA_EXTENSION );  // use the uncompressed extension for md5sum filename

   m_fpa = fopen(m_char_filename_a,"w");
   fprintf(stdout,"filewrite: (file_open) Opening a new file (HK): %s\n", m_char_filename_a );
   
   if ( NULL == m_fpa )  // we had a problem opening the file!?
   {
      printf("Error opening %s\n",m_char_filename_a);
      return 1;
   }
   else               // OK to proceed, reset some variables
   {
      m_counter_fileNumber++;
      m_flag_hk_fileOpen = true;
      m_counter_bytesWritten = 0;  // clear bytes written
      m_time_fileOpenSeconds = time_getEpochSeconds( );
/*
 uint16_t gps_week;
  uint32_t gps_sec_of_week;
  float rtd_temperatures[14];
  float laser_temperatures[3];
  uint8_t laser_status;
  float voltages[3];
  float currents[4];
  float pressures[2];
  float attenuators[2];
  uint32_t hd_space_total[2];
  uint32_t hd_space_used[2];
  int syncClockStatus;
  bool gpsSync;
  float spectrum_temperatures[10];
  float cpu_temperatures[2];
  char currentDataFile[256];
*/
      fprintf( m_fpa,"gps_sec_of_day,rtd_temp,rtd_temp,rtd_temp,rtd_temp,rtd_temp,rtd_temp,rtd_temp,rtd_temp, \
rtd_temp,rtd_temp,rtd_temp,rtd_temp,rtd_temp,rtd_temp,laser_temp,laser_temp,laser_temp,\
lasert_status,voltage,voltage,voltage,current,current,current,current,pressure,pressure,atten,atten,\
hd_total,hd_total,hd_used,hd_used,syncClock Status,gpsSync,spectrum temp, spectrum temp, spectrum temp,\
spectrum temp,spectrum temp,spectrum temp,spectrum temp,spectrum temp,spectrum temp,spectrum temp,\
cpu_temp,cpu_temp,Data File\n"); 
      
      return 0;
   }
}

int lvisf_file_write::file_gps_open( double gpsSeconds )
{
   struct tm myTm;
   time_t local_time;
   double dHour, dMin, dSec;
   int iHour,iMin,iSec;
   
   file_close( );
         
   //
   // Set the day value upon first file open
   //
   if (theDay <= 0)
   {
      time(&local_time);
      memcpy(&myTm,gmtime(&local_time),sizeof(struct tm));
      theYear = myTm.tm_year + 1900;
      theMonth = myTm.tm_mon + 1;
      theDay = myTm.tm_mday;
   }
        
   dHour = gpsSeconds / 3600.0;
   iHour = (int)dHour;
   dMin = dHour - iHour;
   dMin *= 60;
   iMin = (int)dMin;
   dSec = dMin - iMin;
   dSec *= 60;
   iSec = (int)dSec;

   theHour   = myTm.tm_hour = iHour;
   theMinute = myTm.tm_min  = iMin;
   theSecond = myTm.tm_sec  = iSec;
   
   // generate a new file name (first with no path for the md5sum name)
   sprintf( m_char_filename_a, "%s_%04d%02d%02d_%02d%02d%02d_%8.8d", m_char_filename,
	theYear , theMonth , theDay ,
	    myTm.tm_hour , myTm.tm_min , myTm.tm_sec ,
	    m_counter_fileNumber );   
   //printf("Calling file_open [%f] [%.2d:%.2d:%.2d] [%s]\n",gpsSeconds,iHour,iMin,iSec,m_char_filename_a);
   
   strncpy( m_char_filenameUncompressed, m_char_filename_a, sizeof(m_char_filenameUncompressed) );  // copy this file base to here
   // generate a new file name
   sprintf( m_char_filename_a, "%s%s_%04d%02d%02d_%02d%02d%02d_%8.8d", m_char_basepath_a, m_char_filename,
	theYear , theMonth , theDay ,
	    myTm.tm_hour , myTm.tm_min , myTm.tm_sec ,
	    m_counter_fileNumber );
   strcat( m_char_filename_a, m_char_extension );            // add the proper extension
   strcat( m_char_filenameUncompressed, LVISF_DATA_EXTENSION );  // use the uncompressed extension for md5sum filename

   m_fpa = fopen(m_char_filename_a,"ab");
   fprintf(stdout,"filewrite: (file_open) Opening a new file (A): %s\n", m_char_filename_a );

   if( (true == m_flag_secondFileEnabled) && ( strncmp(m_char_basepath_b,"NONE",4)!=0 ) )
     {
	// generate a new file name (first with no path for the md5sum name)
	sprintf( m_char_filename_b, "%s_%04d%02d%02d_%02d%02d%02d_%8.8d", m_char_filename,
	theYear , theMonth , theDay ,
		 myTm.tm_hour , myTm.tm_min , myTm.tm_sec ,
		 m_counter_fileNumber );
	strncpy( m_char_filenameUncompressed, m_char_filename_b, sizeof(m_char_filenameUncompressed) );  // copy this file base to here
	// generate a new file name
	sprintf( m_char_filename_b, "%s%s_%04d%02d%02d_%02d%02d%02d_%8.8d", m_char_basepath_b, m_char_filename,
	theYear , theMonth , theDay ,
		 myTm.tm_hour , myTm.tm_min , myTm.tm_sec ,
		 m_counter_fileNumber );
	strcat( m_char_filename_b, m_char_extension );            // add the proper extension
	strcat( m_char_filenameUncompressed, LVISF_DATA_EXTENSION );  // use the uncompressed extension for md5sum filename

	fprintf(stdout,"filewrite: (file_open) Opening a new file (B): %s\n", m_char_filename_b );

	m_fpb = fopen(m_char_filename_b,"ab");
	if ( NULL == m_fpb )
	  { 
	     fprintf(stderr,"ERROR: Opening file %s \n",m_char_filename_b);
	  }
     }
   
   if ( NULL == m_fpa )  // we had a problem opening the file!?
   {
      printf("Error opening %s\n",m_char_filename_a);
     return 1;
  }
   else               // OK to proceed, reset some variables
     {
	m_counter_fileNumber++;
	m_flag_fileOpen = true;
	m_counter_bytesWritten = 0;  // clear bytes written
	m_time_fileOpenSeconds = time_getEpochSeconds( );
	if( true == m_flag_md5sumFile) MD5_Init(&m_mdContext);
	//if( true == m_flag_compressFile) gzip_init( );
	return 0;
     }
}

void lvisf_file_write::file_set_extension( void )
{
   if ( true == m_flag_compressFile )
     strncpy(m_char_extension,LVISF_DATA_EXTENSION_COMPRESSED,sizeof(m_char_extension));   
   
   if ( false == m_flag_compressFile )
     strncpy(m_char_extension,LVISF_DATA_EXTENSION,sizeof(m_char_extension));

   printf("init - theDay [%d] [%d] [%d]\n",theYear, theMonth, theDay);
}

int lvisf_file_write::get_limit_bytes( void )
{
   return m_limit_bytes;
}

int lvisf_file_write::get_limit_seconds( void )
{
   return m_limit_seconds;
}

int lvisf_file_write::get_var_bytesWritten( void )
{
   return m_counter_bytesWritten;
}

bool lvisf_file_write::get_var_compression( void )
{
   return m_flag_compressFile;
}

bool lvisf_file_write::get_var_md5sum( void )
{
   return m_flag_md5sumFile;
}

/*
void lvisf_file_write::gzip_init ( void )
{
   int ret;
   
   m_strm.zalloc = Z_NULL;
   m_strm.zfree = Z_NULL;
   m_strm.opaque = Z_NULL;
   
   ret = deflateInit2 ( &m_strm,
			LVISF_COMPRESSION_LEVEL,
			Z_DEFLATED,
			LVISF_windowBits | LVISF_GZIP_ENCODING,
			8,        
			Z_DEFAULT_STRATEGY);
   
   if ( Z_OK != ret )
     m_flag_gzipReady = false;
   else
     m_flag_gzipReady = true;
}
*/

// Dan mentions that this could be a "vector" so you don't have to pass size
// but I'm old school and am fine with sending a "length" of the data also
int lvisf_file_write::output_to_file( unsigned char * data , size_t byteSize, bool useInternalBuffer, bool useBaseFilename )
{   
   int retValue = 0;
   //int ret;
   size_t bytesFromFwrite;

   m_counter_blockNumber++;
	
   if((useInternalBuffer == true) && ( byteSize > LVISF_RAW_DATA_SIZE ))
     {
	fprintf(stderr,"ERROR!  Cannot write a block of %d bytes when internal buffer is only %d bytes!?)\n", byteSize ,LVISF_RAW_DATA_SIZE);
	return 404;
     }  
   
   if( true != m_flag_fileOpen ) // if we don't have an open file, create one
     {	
	retValue = file_open( useBaseFilename );  // open a new file
     }
   if( (m_counter_bytesWritten > m_limit_bytes) && (m_limit_bytes > 0) ) // if we've exceeded our file byte size, open a new file
     {
	if( m_limit_bytes > 0 ) // only attempt this if the limit is greater than zero
	  {	     
	     retValue = file_open( useBaseFilename );  // open a new file
	  }
     }
   if( (( time_getEpochSeconds( ) - m_time_fileOpenSeconds ) > m_limit_seconds )   && (m_limit_seconds > 0) )// if we have exceeded our time limits, open a new one
     {
	if( m_limit_seconds > 0 ) // only attempt this if the limit is greater than zero
	  {	     
	     retValue = file_open( useBaseFilename );  // open a new file
	  }	
     }
   if( true == m_flag_fileOpen )  // if we have a file open, write out data to it
     {
	 if(true == useInternalBuffer)
         {
            if( false == m_flag_compressFile)
            {
                memcpy(m_data_out [ m_counter_currentBuffer ], data, byteSize);
                m_counter_lastBuffer = m_counter_currentBuffer;
                m_counter_currentBuffer++;
                if ( m_counter_currentBuffer >= LVISF_RAW_DATA_BUFFERS ) m_counter_currentBuffer = 0;
                bytesFromFwrite = fwrite( m_data_out [ m_counter_lastBuffer ] , 1, byteSize, m_fpa);
                if( bytesFromFwrite > 0 ) m_counter_bytesWritten += ( int ) bytesFromFwrite;
            
            
                if( true == m_flag_secondFileEnabled)
                {
                    if( NULL != m_fpb )
                    {
                        fwrite( m_data_out [ m_counter_lastBuffer ] , 1, byteSize, m_fpb);
                    }
                }
            }
         }
         else
         {
             bytesFromFwrite = fwrite( data , 1, byteSize, m_fpa);
             if( bytesFromFwrite > 0 ) m_counter_bytesWritten += ( int ) bytesFromFwrite;
             
             if( true == m_flag_secondFileEnabled)
             {
                if( NULL != m_fpb )
                {
                    fwrite( data , 1, byteSize, m_fpb);
                }
             }
             
         }
     /*
	if( true == m_flag_compressFile)  // FIXME!?
	  {
	     m_strm.next_in = data;
	     m_strm.avail_in = byteSize;
	     m_strm.avail_out = LVISF_RAW_DATA_SIZE;
	     m_strm.next_out = m_data_out [ m_counter_currentBuffer ];
	     ret = deflate( &m_strm, Z_FULL_FLUSH );
	     //fprintf(stdout,"Return from deflate is %d\n",ret);
	     m_counter_gzipBytesOut =  LVISF_RAW_DATA_SIZE - m_strm.avail_out;
	     bytesFromFwrite = fwrite( m_data_out [ m_counter_currentBuffer ], 1, m_counter_gzipBytesOut, m_fpa);	     
	     if( true == m_flag_secondFileEnabled)
	       if( NULL != m_fpb )
		 {
		    fwrite( m_data_out [ m_counter_currentBuffer ] , 1, m_counter_gzipBytesOut, m_fpb);		    
		 }
	     //fprintf(stdout,"gzip wrote out %d bytes.\n", ( int ) bytesFromFwrite);
	     if ( bytesFromFwrite > 0 ) m_counter_bytesWritten += ( int ) byteSize; // uncompressed bytes written
	  }
	*/
   
	// add to our md5sum
	if( true == m_flag_md5sumFile) MD5_Update (&m_mdContext, (unsigned char *) data, byteSize );
     }

   return retValue;
}

// Dan mentions that this could be a "vector" so you don't have to pass size
// but I'm old school and am fine with sending a "length" of the data also
int lvisf_file_write::output_merged_to_file( unsigned char * data , size_t byteSize, int *fRet1, int *fRet2, double *writeTimeA, double * writeTimeB )
{   
   double tempTime;
   int retValue = 0;
   //int ret;
   size_t bytesFromFwrite;

   m_counter_blockNumber++;
	
   if( byteSize > LVISF_RAW_DATA_SIZE )
     {
	fprintf(stderr,"ERROR!  Cannot write a block of %d bytes when internal buffer is only %d bytes!?)\n", byteSize ,LVISF_RAW_DATA_SIZE);
	return 404;
     }  
   
   if( true == m_flag_fileOpen )  // if we have a file open, write out data to it
     {
	if( true == m_flag_primaryFileEnabled && false == m_flag_compressFile)
	  {
        *fRet1 = *fRet2 = 0;
        
	     /*memcpy(m_data_out [ m_counter_currentBuffer ], data, byteSize);
	     m_counter_lastBuffer = m_counter_currentBuffer;
	     m_counter_currentBuffer++;
	     if ( m_counter_currentBuffer >= LVISF_RAW_DATA_BUFFERS ) m_counter_currentBuffer = 0;
	     bytesFromFwrite = fwrite( m_data_out [ m_counter_lastBuffer ] , 1, byteSize, m_fpa);*/

	     tempTime = time_decimalSecondsNow( );
	     bytesFromFwrite = fwrite(data, 1, byteSize, m_fpa);
	     *writeTimeA = time_decimalSecondsNow( ) - tempTime;
        //
        // Is there an error
        //
        if (bytesFromFwrite != byteSize)
        {
            *fRet1 = bytesFromFwrite;
            errCount1++;
            if (errCount1 > 10)
            {
               //
               // Turn off writing to this device
               //
               m_flag_primaryFileEnabled = false;
               errCount2 = 0;
            }
         }

	     if( bytesFromFwrite > 0 ) m_counter_bytesWritten += ( int ) bytesFromFwrite;
        
        
	     if( true == m_flag_secondFileEnabled)
	       {
		  if( NULL != m_fpb )
		    {
		       tempTime = time_decimalSecondsNow( );
		       bytesFromFwrite = fwrite(data, 1, byteSize, m_fpb);
		       *writeTimeB = time_decimalSecondsNow( ) - tempTime;
		       //bytesFromFwrite = fwrite( m_data_out [ m_counter_lastBuffer ] , 1, byteSize, m_fpb);
             
              //
              // Is there an error
              //
              if (bytesFromFwrite != byteSize)
              {
                  *fRet2 = bytesFromFwrite;
                  errCount2++;
                  if (errCount2 > 10)
                  {
                     //
                     // Turn off writing to this device
                     //
                     m_flag_secondFileEnabled = false;
                     errCount2 = 0;
                  }
              }

		    }
	       }
	  }
     /*
	if( true == m_flag_compressFile)  // FIXME!?
	  {
	     m_strm.next_in = data;
	     m_strm.avail_in = byteSize;
	     m_strm.avail_out = LVISF_RAW_DATA_SIZE;
	     m_strm.next_out = m_data_out [ m_counter_currentBuffer ];
	     ret = deflate( &m_strm, Z_FULL_FLUSH );
	     //fprintf(stdout,"Return from deflate is %d\n",ret);
	     m_counter_gzipBytesOut =  LVISF_RAW_DATA_SIZE - m_strm.avail_out;
	     bytesFromFwrite = fwrite( m_data_out [ m_counter_currentBuffer ], 1, m_counter_gzipBytesOut, m_fpa);	     
	     if( true == m_flag_secondFileEnabled)
	       if( NULL != m_fpb )
		 {
		    fwrite( m_data_out [ m_counter_currentBuffer ] , 1, m_counter_gzipBytesOut, m_fpb);		    
		 }
	     //fprintf(stdout,"gzip wrote out %d bytes.\n", ( int ) bytesFromFwrite);
	     if ( bytesFromFwrite > 0 ) m_counter_bytesWritten += ( int ) byteSize; // uncompressed bytes written
	  }
	*/
   
	// add to our md5sum
	if( true == m_flag_md5sumFile) MD5_Update (&m_mdContext, (unsigned char *) data, byteSize );
     }

   return retValue;
}

int lvisf_file_write::output_hk_to_file( lvisf_housekeeping *hk )
{   
   int retValue = 0;
   size_t bytesFromFwrite;

   m_counter_hk_fileNumber++;
	
   if( (true != m_flag_hk_fileOpen)) // if we don't have an open file, create one
   {	
	   retValue = file_hk_open( );  // open a new house keeping file
   }

   if( m_counter_bytesWritten > m_limit_bytes ) // if we've exceeded our file byte size, open a new file
   {
	   retValue = file_hk_open( );  // open a new file
   }

//   if( ( time_getEpochSeconds( ) - m_time_fileOpenSeconds ) > m_limit_seconds ) // if we have exceeded our time limits, open a new one
//   {
//	   retValue = file_hk_open( );  // open a new file
//   }

   if( true == m_flag_hk_fileOpen && (m_fpa != NULL) )  // if we have a file open, write out data to it
     {
	bytesFromFwrite = fprintf( m_fpa,"%.3lf,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%d,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%d,%d,%d,%d,%d,%d,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%f,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%s\n", 
				   hk->gps_sec_of_the_day,
				   hk->rtd_temperatures[0],
				   hk->rtd_temperatures[1],
				   hk->rtd_temperatures[2],
				   hk->rtd_temperatures[3],
				   hk->rtd_temperatures[4],
				   hk->rtd_temperatures[5],
				   hk->rtd_temperatures[6],
				   hk->rtd_temperatures[7],
				   hk->rtd_temperatures[8],
				   hk->rtd_temperatures[9],
				   hk->rtd_temperatures[10],
				   hk->rtd_temperatures[11],
				   hk->rtd_temperatures[12],
				   hk->rtd_temperatures[13],
				   hk->laser_temperatures[0],
				   hk->laser_temperatures[1],
				   hk->laser_temperatures[2],
				   hk->laser_status,
				   hk->voltages[0],
				   hk->voltages[1],
				   hk->voltages[2],
				   hk->currents[0],
				   hk->currents[1],
				   hk->currents[2],
				   hk->currents[3],
				   hk->pressures[0],
				   hk->pressures[1],
				   hk->attenuators[0],
				   hk->attenuators[1],
				   hk->hd_space_total[0],
				   hk->hd_space_total[1],
				   hk->hd_space_used[0],
				   hk->hd_space_used[0],
				   hk->syncClockStatus,
				   hk->gpsSync,
				   hk->spectrum_temperatures[0],
				   hk->spectrum_temperatures[1],
				   hk->spectrum_temperatures[2],
				   hk->spectrum_temperatures[3],
				   hk->spectrum_temperatures[4],
				   hk->spectrum_temperatures[5],
				   hk->spectrum_temperatures[6],
				   hk->spectrum_temperatures[7],
				   hk->spectrum_temperatures[8],
				   hk->spectrum_temperatures[9],
				   hk->cpu_temperatures[0],
				   hk->cpu_temperatures[1],
				   hk->chmean[0][0],
				   hk->chmean[0][1],
				   hk->chmean[1][0],
				   hk->chmean[1][1],
				   hk->chmean[2][0],
				   hk->chmean[2][1],
				   hk->evenOddMeanDelta[0],
				   hk->evenOddMeanDelta[1],
				   hk->evenOddMeanDelta[2],
				   (int) hk->circular_buffer_depths[0],
				   (int) hk->circular_buffer_depths[1],
				   (int) hk->circular_buffer_depths[2],
				   (int) hk->circular_buffer_depths[3],
				   (int) hk->circular_buffer_depths[4],
				   (int) hk->missed_shots[0],
				   (int) hk->missed_shots[1],
				   (int) hk->missed_shots[2],
				   (int) hk->missed_shots[3],
				   (int) hk->missed_shots[4],
				   (int) hk->laser_rep_rate,
				   hk->currentDataFile);
	
        fflush(m_fpa);
	
	if( bytesFromFwrite > 0 ) m_counter_bytesWritten += ( int ) bytesFromFwrite; 
     }

   return retValue;
}


void lvisf_file_write::set_limit_bytes( uint64_t bytesInput )
{
   m_limit_bytes = bytesInput;
}

void lvisf_file_write::set_limit_seconds( int secondsInput )
{
   m_limit_seconds = secondsInput;
}

void lvisf_file_write::set_var_compression( bool compressionInput )
{
   // check if we already have a file open, if so just close it
   if( true == m_flag_fileOpen )
     file_close( );
   
   m_flag_compressFile = compressionInput;
   file_set_extension( );
}

void lvisf_file_write::set_var_md5sum( bool md5sumInput )
{
   // check if we already have a file open, if so just close it
   if( true == m_flag_fileOpen )
     file_close( );

   m_flag_md5sumFile = md5sumInput;
}

void lvisf_file_write::set_var_md5sum_filename( char * md5filename )
{
   size_t locallength;
   
   // make sure we don't write more than we can handle
   locallength = strlen(md5filename);
   if(locallength > sizeof(m_char_md5sumkernel)) locallength = sizeof(m_char_md5sumkernel);

   memset(m_char_md5sumkernel,0,sizeof(m_char_md5sumkernel));
   strncpy(m_char_md5sumkernel,md5filename,locallength);
   fprintf(stdout,"filewrite: (m_char_md5sumkernel) Setting md5filename to : %s\n", m_char_md5sumkernel );
}

void lvisf_file_write::set_var_path_a( char * pathInput )
{
   size_t locallength;
   
   // make sure we don't write more than we can handle
   locallength = strlen(pathInput);
   if(locallength > sizeof(m_char_basepath_a)) locallength = sizeof(m_char_basepath_a);

   // verify this is not the same path we already have set in B
   if( strncmp( pathInput, m_char_basepath_b, locallength ) != 0 )
     {	
	memset(m_char_basepath_a,0,sizeof(m_char_basepath_a));
	strncpy(m_char_basepath_a,pathInput,locallength);
	fprintf(stdout,"filewrite: (set_var_path_a) Setting path A to : %s\n", m_char_basepath_a );
     }   
}

void lvisf_file_write::set_var_path_b( char * pathInput )
{
   size_t locallength;
   
   // make sure we don't write more than we can handle
   locallength = strlen(pathInput);
   if(locallength > sizeof(m_char_basepath_b)) locallength = sizeof(m_char_basepath_b);

   // verify this is not the same path we already have set in A
   if( strncmp( pathInput, m_char_basepath_a, locallength ) != 0 )
     {
	memset(m_char_basepath_b,0,sizeof(m_char_basepath_b));
	strncpy(m_char_basepath_b,pathInput,locallength);
	fprintf(stdout,"filewrite: (set_var_path_b) Setting path B to : %s\n", m_char_basepath_b );
	
	// should probably check if the path exists and not allow invalid path
	m_flag_secondFileEnabled = true;
     }   
}

void lvisf_file_write::set_filename( char * fn )
{
   size_t locallength;
   
   locallength = strlen( fn );
   if(locallength > sizeof(m_char_filename)) locallength = sizeof(m_char_filename);
   
   memset(m_char_filename,0,sizeof(m_char_filename));
   strncpy(m_char_filename,fn,locallength);
   fprintf(stdout,"filewrite: (set_filename) Setting filename to : %s\n", m_char_filename );
}

char *lvisf_file_write::get_filename( char * fn, int whichFlag, size_t maxsize )
{
   if (whichFlag == 1)
      strncpy(fn, m_char_filename_a, maxsize);
   else
      strncpy(fn, m_char_filename_b, maxsize);
   
   return fn;
   //fprintf(stdout,"filewrite: (set_filename) Setting filename to : %s\n", m_char_filename );
}

int lvisf_file_write::time_getEpochSeconds( void )
{
   int status;
   struct timeval currentTime;
   struct timezone tzp;
   
   status = gettimeofday(&currentTime,&tzp);
   
   return (int) currentTime.tv_sec;
}

double lvisf_file_write::time_decimalSecondsNow( void )
{
   int status;
   struct timespec currentTime;
   
   status = clock_gettime(CLOCK_MONOTONIC_RAW, &currentTime);
   if(status)
     {
	
	// handle error
	return 0.0;
     }
   else
     {
	return ( (double) currentTime.tv_sec + ((double) currentTime.tv_nsec) / 1000000000.0 );
     }
}

double lvisf_file_write::getDiskStatus()
{   
   struct statvfs statvbuf0;
   double available;
   char fn[512];
  
   //
   // Filenames and path
   //
   memset(fn,'\0',sizeof(fn));
   strncpy(fn,"/mnt/data/",sizeof(fn));

   //printf("[%s] [%s]\n",fn,fn);

   //
   // available
   //
   if (statvfs(fn, &statvbuf0) != 0)
   {
      printf("statvfs Error\n");
   }
   else
   {
      available = (statvbuf0.f_bsize * statvbuf0.f_bavail) / 1000000000;
      printf("getDiskStatus [%f]\n",available);
      
      return available;
   }
    
    return 0;
}

bool lvisf_file_write::isOpen()
{
   return m_flag_fileOpen;   
}


uint32_t lvisf_file_write::get_yyyymmdd()
{
   return (theYear *10000) + (theMonth * 100) + theDay;
}

uint32_t lvisf_file_write::get_hhmmss()
{
   return (theHour *10000) + (theMinute * 100) + theSecond;
}

uint32_t lvisf_file_write::get_filenumber()
{
   return m_counter_fileNumber;
}
