#ifndef _LVISF_FILEWRITECLASS_H_
#define _LVISF_FILEWRITECLASS_H_

#include <openssl/md5.h>  // guess what needs this!?
#include <zlib.h>     // required by gzip
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "lvisf_structures.h"

// compression hard coded options
#define LVISF_COMPRESSION_LEVEL Z_BEST_SPEED
#define LVISF_GZIP_ENCODING 16
#define LVISF_windowBits 15


const char LVISF_DATA_BASEPATH[] = "/mnt/data/";
const char LVISF_DATA_EXTENSION[] = ".dat";
const char LVISF_DATA_EXTENSION_COMPRESSED[] = ".dat.gz";
const char LVISF_FILENAME_NOT_SET[] = "filename_not_set";
const char LVISF_FILENAME_NONE[] = "NONE";
const char LVISF_MD5SUM_DEFAULT[] = "md5deep_rtlvis.txt";

// internal buffers
#define LVISF_RAW_DATA_SIZE ( 100 * 1024 * 1024 )  // 10 MB (so 1000 x 9k is OK!)
#define LVISF_RAW_DATA_BUFFERS 2            // number of buffers

class lvisf_file_write
{

 public:
   lvisf_file_write();
   ~lvisf_file_write();
   int calc_file_md5sum( char * filename, unsigned char * md5result );
   int get_file_modtime( char * filename, char * datestring );
   int get_limit_bytes( void );
   int get_limit_seconds( void );
   int get_var_bytesWritten( void );
   bool get_var_compression( void );
   bool get_var_md5sum( void );
   int output_to_file( unsigned char * , size_t, bool useInternalBuffer = true, bool useBaseFilename = false );
   int output_merged_to_file( unsigned char * data , size_t byteSize, int *fRet1, int *fRet2, double *writeTimeA, double *writeTimeB );
   int output_hk_to_file( lvisf_housekeeping *hk );
   void set_limit_bytes( uint64_t );
   void set_limit_seconds( int );
   void set_var_compression( bool );
   void set_var_md5sum( bool );
   void set_var_md5sum_filename( char * );
   void set_var_path_a( char * );
   void set_var_path_b( char * );
   void set_filename( char * );
   char  *get_filename( char * fn, int whichFlag , size_t maxsize);
   uint32_t get_yyyymmdd();
   uint32_t get_hhmmss();
   uint32_t get_filenumber();
   
   double getDiskStatus();
   bool isOpen();
   void file_close( void );
   void file_flushCache( void );
   void print_theDay();

   int file_gps_open( double gpsSeconds );
   void init();

 private:  // private functions
   int file_open( bool useBaseFilename );
   int file_hk_open( void );
   void file_set_extension( void );
   //void gzip_init( void );
   int time_getEpochSeconds( void );
   double time_decimalSecondsNow( void );
   
   // private variable are usually started with m_ or p_ (for mine or private respectively)
 private:

   char m_char_extension[32];
   char m_char_filename[2048];
   char m_char_basepath_a[2048];
   char m_char_basepath_b[2048];
   char m_char_filename_a[2048];
   char m_char_filename_b[2048];
   char m_char_md5sumkernel[2048];        // actual name used for md5sum file
   char m_char_filename_md5sum[2048];     // used to create the filename
   char m_char_filenameUncompressed[2048];
   int m_counter_blockNumber;
   uint64_t m_counter_bytesWritten;
   int m_counter_fileNumber;
   int m_counter_hk_fileNumber;
   int m_counter_gzipBytesOut;
   int m_counter_currentBuffer;
   int m_counter_lastBuffer;
   unsigned char * m_data_out[LVISF_RAW_DATA_BUFFERS];  // where our compressed data will end up
   bool m_flag_compressFile;
   bool m_flag_fileOpen;
   bool m_flag_hk_fileOpen;
   bool m_flag_gzipReady;
   bool m_flag_md5sumFile;
   bool m_flag_openNewFile;
   bool m_flag_primaryFileEnabled;
   bool m_flag_secondFileEnabled;
   bool m_flag_systemInitialized;
   uint64_t m_limit_bytes;
   int m_limit_bytesPerWrite;  // how many bytes at a time to write
   int m_limit_delayPerWrite;  // how many useconds to wait between writes
   int m_limit_seconds;
   int m_time_fileOpenSeconds;
   MD5_CTX m_mdContext;
   unsigned char m_md5sum[MD5_DIGEST_LENGTH];
   FILE * m_fp, * m_fpa, * m_fpb;
   int m_fid;
   //z_stream m_strm;
   int errCount1, errCount2;

   //
   // date values to save the current date
   // upon the first file open. Will be used on all future
   // file opens - do not want date to change
   //
   int theDay;
   int theMonth;
   int theYear;

   int theHour;
   int theMinute;
   int theSecond;

};

#endif
